#!/usr/bin/perl
#
# RDF::Parser module
#
# (C)opyright 1999 Pro Solutions Ltd. <info@pro-solutions.com>
#
# This module is used to compile an XML/RDF description into
# the triple representation of the underlying RDF data model.
#
# Please read License.txt for licensing.
#
# This software conforms to the
#
# Resource Description Framework (W3C) Model & Syntax specification
# a W3C Recommendation dated February 22nd, 1999 available at
# http://www.w3.org/TR/1999/REC-rdf-syntax-19990222/
# 
# with the exception that the 'aboutEachPrefix' feature is not
# supported.
#
package RDF::Parser;
use vars qw($VERSION);
use Carp;
use strict;
use XML::Parser;

$VERSION = '1.01';

$::RDFSCHEMA    = "http://www.w3.org/1999/02/22-rdf-syntax-ns#";
$::RDF_triples  = ();
$::RDF_parser   = undef;
@::elementStack = ();

sub new
  {
    my ($class)		= shift;
    my ($self)		= {};

    #
    $self->{TELLABLES}	= ();

    $self->{BASEURI}	= undef;

    bless $self, $class;
  }

sub base
  {
    my ($self) = shift;
    my ($uri)  = shift;

    if (defined $uri) {
      $self->{BASEURI} = $uri;
    } else {
      return $self->{BASEURI};
    }
  }

sub parsefile
  {
    my ($self)    = shift;
    my ($file)    = shift;
    my ($content) = "";

    open (FILE, "$file") || die "Cannot open $file\n";
    $/ = undef;
    $content = <FILE>; 
    close (FILE);

    $self->parse ($content);
  }

sub parse
  {
    my ($self)    = shift;
    my ($content) = shift;
    $::RDF_parser = $self;

    $self->{ROOT}	= undef;

    $self->{TRIPLES}	= ();

    $self->{ERRORS}	= ();

    $self->{LITERAL}	= "";

    $self->{PARSETYPESTACK} = ();

    $self->{PCDATA}	= undef;

    $self->{IDS}     = 0;
    $self->{PACKAGE} = caller;

    my $xmlparser = new XML::Parser (Handlers => {Start => \&startElement,
						  End   => \&endElement,
						  Final => \&startRDF,
						  Char  => \&characters},
				     Namespaces => 1);
    foreach my $tellable (@ { $self->{TELLABLES}}) {
      my $start = $$tellable[0];
      &$start ($self);
    }

    eval { $xmlparser->parse ($content) };
    if ($@) {
      $self->addError ("XML::Parser reported an error '" . $@ . "'");
    }

    foreach my $tellable (@ { $self->{TELLABLES}}) {
      my $end = $$tellable[1];
      &$end ($self);
    }
  }

sub startElement
  {
    my ($self)		= $::RDF_parser;
    my ($expat)         = shift;
    my ($element)       = shift;
    my (@attributelist) = @_;
    my ($ns)            = $expat->namespace($element);
    my (%attributes)	= ();

    if ($ns eq "" && defined $self->{BASEURI}) {
      $ns = $self->{BASEURI};
    }

    $self->{PCDATA}	= undef;

    for (my $x = 0; $x < $#attributelist; $x += 2) {
      my ($attributename)  = $attributelist[$x];
      my ($attributevalue) = $attributelist[$x+1];
      my ($attributens)    = $expat->namespace ($attributename);
      if (!(defined $attributens) &&
	  !($attributename =~ /^xml:/)) {
	$attributens = $ns;
      }

      $attributes{$attributens . $attributename} = $attributevalue;
    }

    my $newElement = RDF::Element->new ($ns, $element,
					\%attributes);

    my $id = $newElement->attribute ($::RDFSCHEMA . "ID");
    if (defined $id) {
      if (defined $self->{BASEURI}) {
	my $uri = URI->new_abs ($id, $self->{BASEURI});
	$id = $uri->as_string;
      }
      $self->registerNode ($id, $newElement);
      $newElement->ID ($id);
    }
    
    my $bagid = $newElement->attribute ($::RDFSCHEMA . "bagID");
    if (defined $bagid) {
      if (defined $self->{BASEURI}) {
	my $uri = URI->new_abs ($bagid, $self->{BASEURI});
	$bagid = $uri->as_string;
      }
      $self->registerNode ($bagid, $newElement);
      $newElement->bagID ($bagid);
    }
    
    my $language = $newElement->attribute ("xml:lang");
    if (!(defined $language)) {
      my $parent = pop @::elementStack;
      if (defined $parent) {
	push @::elementStack, $parent;
	$newElement->attribute ("xml:lang",
				$parent->attribute("xml:lang"));
      }
    }

    my $parseType = $newElement->attribute ($::RDFSCHEMA . "parseType");
    if (defined $parseType &&
	!($self->parsingLiteral()) &&
	$parseType ne "Resource") {

      push @ { $self->{PARSETYPESTACK} }, $newElement;

      my $parent = pop @::elementStack;
      if (defined $parent) {
	push @::elementStack, $parent;
	$parent->addChild($newElement),
      }

      push @::elementStack, $newElement;

      return;
    }

    if ($self->parsingLiteral()) {
      $self->makeST ($newElement);
      push @::elementStack, $newElement;
      return;
    }

    my $parent = pop @::elementStack;
    if (defined $parent) {
      push @::elementStack, $parent;
      $parent->addChild($newElement),
    }

    push @::elementStack, $newElement;

    if (defined $parseType &&
	$parseType eq "Resource") {
      push @ { $self->{PARSETYPESTACK} }, $newElement;
      my $desc = RDF::Element->new ($::RDFSCHEMA, "Description",
				    undef);

      $desc->generated (1);
      if (defined $::elementStack[$#::elementStack]) {
	my $parent = $::elementStack[$#::elementStack];
	$parent->addChild ($desc);
	push @::elementStack, $desc;
      }
    }
  }

sub endElement
  {
    my ($self)		  = $::RDF_parser;
    my ($expat, $element) = @_;
    my ($parseLiteral)    = $self->parsingLiteral();
    $self->{ROOT} = pop @::elementStack;

    $self->{PCDATA}	= undef;

    if ($self->{ROOT}->generated()) {
      $self->{ROOT} = pop @::elementStack;
    }

    if ($parseLiteral) {
      my $scopeElement = pop @ { $self->{PARSETYPESTACK} };
      if ($scopeElement == $self->{ROOT}) {
	$self->{ROOT}->addChild (RDF::Data->new ($self->{LITERAL}));
	$self->{LITERAL} = "";
      } else {
	push @ { $self->{PARSETYPESTACK} }, $scopeElement;
	$self->makeET ($self->{ROOT});
      }
    }
  }

sub characters
  {
    my ($self)		 = $::RDF_parser;
    my ($expat, $string) = @_;

    if ($self->parsingLiteral()) {
      $self->makeContent ($string);
      return;
    }

    my $parent = $::elementStack[$#::elementStack];
    my $parseType = $parent->attribute ($::RDFSCHEMA . "parseType");
    if (defined $parseType &&
	$parseType eq "Resource") {
      return;
    }

    if (defined $self->{PCDATA}) {
      my $olddata = $self->{PCDATA}->data();
      $self->{PCDATA}->data ($olddata . $string);
    } else {
      my $data = RDF::Data->new ($string);
      $self->{PCDATA} = $data;
      my ($parent) = pop @::elementStack;
      if (defined $parent) {
	push @::elementStack, $parent;
	$parent->addChild ($data);
      }
    }
  }

sub registerNode
  {
    my ($self)		= shift;
    my ($id)		= shift;
    my ($node)		= shift;

    if (defined $self->findNode ($id)) {
    } else {
      $self->{NODES}{$id} = $node;
    }
  }

sub findNode
  {
    my ($self)		= shift;
    my ($id)		= shift;

    return $self->{NODES}{$id};
  }

sub makeST
  {
    my ($self)		= shift;
    my ($element)	= shift;

    my ($genid)         = undef;

    do {
      $genid = $self->{IDS}++;
    } while (defined $element->attribute ("xmlns:" . $genid));

    $element->prefix ("genid" . $genid);

    $self->{LITERAL} .= "<genid" . $genid . ":" . $element->local();
    $self->{LITERAL} .= " xmlns:genid" . $genid . "=\"" . $element->namespace() . "\"";
    foreach my $attributename (keys % { $element->{ATTRIBUTES} } ) {
      my $attributevalue = $element->{ATTRIBUTES}{$attributename};
      $self->{LITERAL} .= " $attributename='$attributevalue'";
    }
    $self->{LITERAL} .= ">";
  }

sub makeET
  {
    my ($self)		= shift;
    my ($element)	= shift;

    $self->{LITERAL} .= "</" . $element->prefix() . ":" . $element->local() . ">";
  }

sub makeContent
  {
    my ($self)		= shift;
    my ($chars)		= shift;

    $self->{LITERAL} .= $chars;
  }

sub parsingLiteral
  {
    my ($self) = shift;

    for (my $x = $#::elementStack; $x >= 0; $x--) {
      my $element = $::elementStack[$x];
      my $parseLiteral = $element->attribute ($::RDFSCHEMA . "parseType");
      if (defined $parseLiteral) {
	if ($parseLiteral ne "Resource") {
	  return 1;
	}
      }
    }
    return 0;
  }

sub parsingResource
  {
    my ($self) = shift;

    for (my $x = $#::elementStack; $x >= 0; $x--) {
      my $element = $::elementStack[$x];
      my $parseType = $element->attribute ($::RDFSCHEMA . "parseType");
      if (defined $parseType) {
	if ($parseType eq "Resource") {
	  return 1;
	}
      }
    }
    return 0;
  }

sub triples
  {
    my ($self)  = shift;
    my (@empty) = ();

    if (defined $self->{TRIPLES}) {
      return @ { $self->{TRIPLES} };
    } else {
      return @empty;
    }
  }

sub newID
  {
    my ($self)    = shift;
    my ($newid)   = "genid" . $self->{IDS}++;

    if (defined $self->{BASEURI}) {
      my $uri = URI->new_abs ($newid, $self->{BASEURI});
      $newid = $uri->as_string;
    }
    return $newid;
  }

sub startRDF
  {
    my ($self)		= $::RDF_parser;
    my ($expat)		= shift;


    my @children = $self->{ROOT}->children();
    foreach my $child (@children) {
      if ($self->isDescription($child)) {
	$self->doDescription ($child);
      } elsif ($self->isData($child)) {
      } elsif ($self->isContainer($child)) {
	$self->doContainer($child);
      } else {
	$self->doTypedNode($child);
      }
    }

    return 1;
  }

sub doDescription
  {
    my ($self) = shift;
    my ($description) = shift;

    if ($description->done()) {
      return;
    } else {
      $description->done (1);
    }

    my ($about)		= $description->attribute ($::RDFSCHEMA . "about");
    my ($aboutEach)	= $description->attribute ($::RDFSCHEMA . "aboutEach");
    my ($id)		= $description->ID();
    my ($bagid) 	= $description->bagID();

    if (defined $about && $about eq "") {
      $about = $self->{BASEURI};
    }


    if (defined $about && defined $id) {
      $self->addError ("Description cannot have both 'about' ".
		       "and 'ID' attributes defined");
    }

    my $childcounter = 1;
    if (defined $bagid) {
      $self->addTriple (RDF::Resource->new ($bagid),
			RDF::Predicate->new ($::RDFSCHEMA . "type"),
			RDF::Resource->new ($::RDFSCHEMA . "Bag"));
    }
    if (!(defined $id)) {
      $id = $about;
    }
    if (!(defined $id)) {
      $id = $self->newID();
    }
    my ($subject) = RDF::Resource->new ($id);

    $description->expandAttributes;

    if (defined $aboutEach) {
      $aboutEach =~ s/^#//;
      if (defined $self->{BASEURI}) {
	my $uri = URI->new_abs ($aboutEach, $self->{BASEURI});
	$aboutEach = $uri->as_string;
      }

      my $target = $self->findNode ($aboutEach);

      if ($self->isDescription ($target)) {
	$self->doDescription ($target)
      } else {
	$self->doTypedNode ($target);
      }

      if (defined $target) {
	my @targetchildren = $target->children();
	foreach my $targetchild (@targetchildren) {
	  next if ($self->isData ($targetchild));

	  my $newDescription = RDF::Element->new ($::RDFSCHEMA,
						  "Description",
						  undef);
	  $newDescription->ID ($targetchild->ID());
	  my @descriptionchildren = $description->children();
	  foreach my $child (@descriptionchildren) {
	    $newDescription->addChild ($child);
	  }
	  $self->doDescription ($newDescription);
	}
	return undef;
      } else {
	$self->addError ("Unresolved internal reference to '" .
			 $aboutEach . "'");
      }
    }

    my @children = $description->children();
    foreach my $child (@children) {
      if ($self->isDescription($child)) {
	$self->addError ("Cannot nest Description inside ".
			 "another Description");
      } elsif ($self->isData($child)) {
      } else {
	my ($predicate, $object, $reification_id) = $self->doPredicate($child);
	my $triple = $self->addTriple($subject, $predicate, $object);

	if (defined $bagid) {
	  $child->ID ($reification_id);
	  $self->addTriple (RDF::Resource->new ($bagid),
			    RDF::Predicate->new ($::RDFSCHEMA . "_$childcounter"),
			    RDF::Resource->new ($reification_id));
	  $childcounter++;

	  $self->reificate ($triple, RDF::Resource->new ($reification_id));
	}
      }
    }

    return RDF::Resource->new($id);
  }

sub doPredicateAttributes
  {
    my ($self)		= shift;
    my ($pred)		= shift;
    my ($id)		= $pred->ID();
    my ($foundatt)	= 0;

    if (!(defined $id)) {
      $id = $pred->attribute($::RDFSCHEMA . "resource");
    }
    if (!(defined $id)) {
      $id = $self->newID();
    }

    foreach my $attributename (keys % { $pred->{ATTRIBUTES} } ) {
      if ($attributename ne $::RDFSCHEMA . "ID" &&
	  $attributename ne $::RDFSCHEMA . "resource" &&
	  $attributename ne $::RDFSCHEMA . "about" &&
	  $attributename ne $::RDFSCHEMA . "aboutEach" &&
	  $attributename ne $::RDFSCHEMA . "aboutEachPrefix" &&
	  $attributename ne $::RDFSCHEMA . "bagID" &&
	  $attributename ne $::RDFSCHEMA . "parseType" &&
	  !($attributename =~ /^xml:/)) {
	my $attributevalue = $pred->{ATTRIBUTES}{$attributename};
	delete $pred->{ATTRIBUTES}{$attributename};

	my $subject	= RDF::Resource->new($id);
	my $predicate	= RDF::Predicate->new ($attributename);
	my $object	= RDF::Literal->new ($attributevalue);

	$self->addTriple ($subject, $predicate, $object);

	$foundatt = 1;
      }
    }
    return $foundatt;
  }

sub doTypedNode
  {
    my ($self) = shift;
    my ($typedNode) = shift;

    my $resource	= $typedNode->attribute ($::RDFSCHEMA . "resource");
    if (defined $resource) {
      $self->addError ("A typedNode '" . $typedNode->name() .
		       "' cannot have rdf:resource attribute");
    }

    my $aboutEach	= $typedNode->attribute ($::RDFSCHEMA . "aboutEach");
    if (defined $aboutEach) {
      my $newPredicate = RDF::Element->new ($::RDFSCHEMA,
					    "type",
					    undef);
      my $newData = RDF::Data->new ($typedNode->name());
      $newPredicate->addChild ($newData);
      $typedNode->addChild ($newPredicate);
      return $self->doDescription ($typedNode);

    } else {
      my $descID        = $self->doDescription ($typedNode);
      my $subject	= RDF::Resource->new($descID->ID());
      my $predicate	= RDF::Predicate->new($::RDFSCHEMA . "type");
      my $object	= RDF::Resource->new($typedNode->name());

      $self->addTriple					     ($subject, $predicate, $object);

      return $subject;
    }
  }

sub doPredicate
  {
    my ($self)		= shift;
    my ($pred)		= shift;
    my ($predicate)	= RDF::Predicate->new($pred->name());
    my ($object)	= undef;
    my ($resource)	= $pred->attribute ($::RDFSCHEMA . "resource");
    my ($reification_id)= $pred->ID();

    if (!(defined $reification_id)) {
      $reification_id = $self->newID();
      $pred->ID ($reification_id);
    }

    my @children = $pred->children();
    if ($self->doPredicateAttributes ($pred)) {
      if ($#children > -1) {
	$self->addError ("Cannot have nested children in ".
			 "production 6.12 with propAttr rule activated");
      }
    }

    if (defined $resource) {
      if ($resource =~ /^#/) {
	  $resource =~ s/^#//;

	  if (defined $self->{BASEURI}) {
	    my $uri = URI->new_abs ($resource, $self->{BASEURI});
	    $resource = $uri->as_string;
          }

          if (!(defined $self->findNode ($resource))) {
	    $self->addError ("Unresolved internal reference to '" .
			     $resource . "'");
	  }
      }
      return ($predicate, RDF::Resource->new ($resource), $reification_id);
    }

    foreach my $child (@children) {
      if ($self->isDescription ($child)) {
	$object = $self->doDescription ($child);
	last; # do not consider CDATA appearing after a Description
      } elsif ($self->isData($child)) {
	$object = RDF::Literal->new ($child->data(),
				     $pred->attribute("xml:lang"));
      } elsif ($self->isContainer($child)) {
	$object = $self->doContainer ($child);
	last; # do not consider CDATA appearing after a Container
      } else {
	$object = $self->doTypedNode ($child);
	last; # do not consider CDATA appearing after the typedNode
      }
    } 

    if (defined $object) {
      return ($predicate, $object, $reification_id);
    } else {
      return ($predicate, RDF::Literal->new (""), $reification_id);
    }
 }

sub doContainer
  {
    my ($self)		= shift;
    my ($container)	= shift;
    my ($li_counter)	= 1;
    my ($id)		= $container->ID();
    if (!(defined $id)) {
       $id = $self->newID();
    }
    my ($subject)	= RDF::Resource->new($id);
    
    $self->addTriple (RDF::Resource->new ($id),
		      RDF::Predicate->new ($::RDFSCHEMA . "type"),
		      RDF::Resource->new ($container->name()));

    my @children = $container->children();
    foreach my $child (@children) {
      if ($self->isListitem ($child)) {
	my $object = $self->doListitem ($child);
	my $predicate = RDF::Predicate->new ($::RDFSCHEMA . "_$li_counter");
	$self->addTriple ($subject, $predicate, $object);
	$li_counter++;
      } elsif ($self->isData($child)) {
      }
    }

    if ($li_counter == 1 &&
	$self->isAlternative ($container)) {
      $self->addError ("An Alt element needs at least one 'rdf:li' child");
    }

    return (RDF::Resource->new($id));
  }

sub doListitem
  {
    my ($self) = shift;
    my ($listitem) = shift;

    my @children = $listitem->children();
    my $removecounter = 0;
    if ($#children > 0) {
      foreach my $child (@children) {
	if ($self->isData ($child)) {
	  $removecounter++;
	} else {
	  last;
	}
      }
    }
    if ($removecounter > 0) {
      $listitem->removeChildren (0, $removecounter);
    }

    my $resource = $listitem->attribute ($::RDFSCHEMA . "resource");
    if (defined $resource) {
      if ($#children > -1) {
	$self->addError ("A listitem with rdf:attribute ".
			 "cannot have child nodes");
      }

      $listitem->ID ($resource);
      return RDF::Resource->new ($resource);
    }

    foreach my $child (@children) {
      if ($self->isDescription ($child)) {
	my $resource = RDF::Resource->new($self->doDescription ($child));
	$listitem->ID ($resource->ID());
	return $resource;
      } elsif ($self->isData($child)) {
	return RDF::Literal->new($child->data(),
				$listitem->attribute("xml:lang"));
      } else {
	my $resource = $self->doTypedNode ($child);
	$listitem->ID ($resource->ID());
	return $resource;
      }
    }
  }

sub addError
  {
    my ($self) = shift;
    my ($msg)  = shift;

    push @ { $self->{ERRORS} }, $msg;

    foreach my $tellable (@ { $self->{TELLABLES}}) {
      my $error = $$tellable[3];
      &$error ($self, $msg);
    }
  }

sub errors
  {
    my ($self) = shift;

    if (defined $self->{ERRORS}) {
      return @ { $self->{ERRORS} };
    } else {
      return;
    }
  }

sub addTriple
  {
    my ($self)		= shift;
    my ($subject)	= shift;
    my ($predicate)	= shift;
    my ($object)	= shift;

    my $triple = RDF::Triple->new ($subject, $predicate, $object);
    push @ { $self->{TRIPLES} }, $triple;

    foreach my $tellable (@ {$self->{TELLABLES}}) {
      my $assert = $$tellable[2];
      &$assert ($self, $triple);
    }

    return $triple;
  }

sub reificate
  {
    my ($self)		= shift;
    my ($triple)	= shift;
    my ($identity)	= shift;

    my ($subject)	= $triple->subject ();
    my ($predicate)	= $triple->predicate ();
    my ($object)	= $triple->object ();

    $self->addTriple ($identity,
		      RDF::Predicate->new ($::RDFSCHEMA . "subject"),
		      RDF::Resource->new($subject->name()));

    $self->addTriple ($identity,
		      RDF::Predicate->new ($::RDFSCHEMA . "predicate"),
		      RDF::Resource->new($predicate->name()));

    $self->addTriple ($identity,
		      RDF::Predicate->new ($::RDFSCHEMA . "object"),
		      RDF::Resource->new($object->name()));

    $self->addTriple ($identity,
		      RDF::Predicate->new ($::RDFSCHEMA . "type"),
		      RDF::Resource->new($::RDFSCHEMA . "Statement"));
  }

sub isData
  {
    my ($self) = shift;
    my ($element) = shift;

    return defined $element->{DATA};
  }

sub isRDF
  {
    my ($self) = shift;
    my ($element) = shift;

    return ($element->name() eq $::RDFSCHEMA . "RDF");
  }

sub isDescription
  {
    my ($self) = shift;
    my ($element) = shift;

    return ($element->name() eq $::RDFSCHEMA . "Description");
 }

sub isSequence
  {
    my ($self) = shift;
    my ($element) = shift;

    return ($element->name() eq $::RDFSCHEMA . "Seq");
  }

sub isBag
  {
    my ($self) = shift;
    my ($element) = shift;

    return ($element->name() eq $::RDFSCHEMA . "Bag");
  }

sub isAlternative
  {
    my ($self) = shift;
    my ($element) = shift;

    return ($element->name() eq $::RDFSCHEMA . "Alt");
  }

sub isContainer
  {
    my ($self)		= shift;
    my ($element)	= shift;

    return ($self->isSequence ($element) ||
	    $self->isBag ($element) ||
	    $self->isAlternative ($element));
  }

sub isListitem
  {
    my ($self)		= shift;
    my ($element)	= shift;

    return ($element->name() eq $::RDFSCHEMA . "li");
  }

sub registerTellable
  {
    my ($self)				= shift;
    my ($start,$end,$assert,$error)	= @_;
    my @items				= ($start, $end, $assert, $error);

    push @ { $self->{TELLABLES}}, \@items;
  }

package RDF::Triple;

sub new
  {
    my ($class) = shift;
    my ($self) = {};

    $self->{SUBJECT}   = shift;
    $self->{PREDICATE} = shift;
    $self->{OBJECT}    = shift;

    bless $self, $class;
  }

sub subject
  {
    my ($self) = shift;

    return $self->{SUBJECT};
  }

sub predicate
  {
    my ($self) = shift;

    return $self->{PREDICATE};
  }

sub object
  {
    my ($self) = shift;

    return $self->{OBJECT};
  }

sub print
  {
    my ($self) = shift;

    print "triple(";
    if (defined $self->{SUBJECT}) {
      print $self->{SUBJECT}->print();
    }
    print ",";
    if (defined $self->{PREDICATE}) {
      print $self->{PREDICATE}->print();
    }
    print ",";
    if (defined $self->{OBJECT}) {
      print $self->{OBJECT}->print();
    }
    print ").\n";
  }

package RDF::RDFNode;

sub new
  {
    my ($class) = shift;
    my ($self) = {};
    $self->{NAME} = shift;

    bless $self, $class;
  }

sub name
  {
    my ($self) = shift;

    return $self->{NAME};
  }

package RDF::Resource;

@RDF::Resource::ISA = qw( RDF::RDFNode );

sub new
  {
    my ($class) = shift;
    my ($name) = shift;
    my ($self) = $class->SUPER::new($name);

    bless $self, $class;
  }

sub print
  {
    my ($self) = shift;

    return "resource('" . $self->name() . "')";
  }

sub ID
  {
    my ($self) = shift;

    return $self->name();
  }

package RDF::Literal;

@RDF::Literal::ISA = qw( RDF::RDFNode );

sub new
  {
    my ($class)		= shift;
    my ($name)		= shift;
    my ($language)	= shift;
    my ($self)		= $class->SUPER::new($name);

    if (defined $language) {
      $self->{LANGUAGE} = $language;
    } else {
      $self->{LANGUAGE} = undef;
    }

    bless $self, $class;
  }

sub language
  {
    my ($self)		= shift;
    my ($language)	= shift;

    if (defined $language) {
      $self->{LANGUAGE} = $language;
    } else {
      return lc $self->{LANGUAGE};
    }
  }

sub print
  {
    my ($self) = shift;

    my $content = $self->name();
    $content =~ s/\\/\\\\/g;
    $content =~ s/'/\\'/g;
    $content =~ s/\n/\\n/g;
    $content =~ s/\r/\\r/g;

    if (defined $self->{LANGUAGE}) {
      return "literal('" . $content . "', '" . $self->language() . "')";
    } else {
      return "literal('" . $content . "')";
    }
  }

package RDF::Predicate;

@RDF::Predicate::ISA = qw( RDF::RDFNode );

sub new
  {
    my ($class) = shift;
    my ($name) = shift;
    my ($self) = $class->SUPER::new($name);

    bless $self, $class;
  }

sub print
  {
    my ($self) = shift;

    return "predicate('" . $self->name() . "')";
  }

package RDF::Element;

sub new
  {
    my ($class)    = shift;
    my ($self) = {};

    $self->{NAMESPACE}	= shift;
    $self->{NAME}	= shift;
    $self->{ATTRIBUTES} = shift;
    $self->{ID}		= undef;
    $self->{BAGID}	= undef;
    $self->{PREFIX}	= undef;
    $self->{CHILDREN}	= ();
    $self->{DONE}	= undef;

    bless $self, $class;
  }

sub name
  {
    my ($self) = shift;

    return $self->{NAMESPACE} . $self->{NAME};
  }

sub local
  {
    my ($self) = shift;

    return $self->{NAME};
  }

sub namespace
  {
    my ($self) = shift;

    return $self->{NAMESPACE};
  }

sub prefix
  {
    my ($self)		= shift;
    my ($prefix)	= shift;

    if (defined $prefix) {
      $self->{PREFIX} = $prefix;
    } else {
      return $self->{PREFIX};
    }
  }

sub done
  {
    my ($self)		= shift;
    my ($flag)		= shift;

    if (defined $flag) {
      $self->{DONE} = $flag;
    } else {
      return $self->{DONE};
    }
  }

sub removeChildren
  {
    my ($self)		= shift;
    my ($start)		= shift;
    my ($length)	= shift;

    splice (@ { $self->{CHILDREN} }, $start, $length);
  }

sub attribute
  {
    my ($self) = shift;
    my ($attname) = shift;
    my ($attvalue) = shift;

    if (defined $attvalue) {
      $self->{ATTRIBUTES}{$attname} = $attvalue;
    } else {
      return $self->{ATTRIBUTES}{$attname};
    }
  }

sub generated
  {
    my ($self)  = shift;
    my ($flag)  = shift;

    if (defined $flag) {
      $self->{GENERATED} = $flag;
    } else {
      return $self->{GENERATED};
    }
  }

sub ID
  {
    my ($self)  = shift;
    my ($value) = shift;

    if (defined $value) {
      $self->{ID} = $value;
    } else {
      return $self->{ID};
    }
  }

sub bagID
  {
    my ($self)  = shift;
    my ($value) = shift;

    if (defined $value) {
      $self->{BAGID} = $value;
    } else {
      return $self->{BAGID};
    }
  }

sub children
  {
    my ($self)  = shift;
    my (@empty) = ();

    if (defined $self->{CHILDREN}) {
      return @ { $self->{CHILDREN} };
    } else {
      return @empty;
    }
  }

sub addChild
  {
    my ($self) = shift;
    my ($element) = shift;

    push @ { $self->{CHILDREN} }, $element;
  }

sub expandAttributes
  {
    my ($self) = shift;

    foreach my $attributename (keys % { $self->{ATTRIBUTES} } ) {
      my $attributevalue = $self->{ATTRIBUTES}{$attributename};

      if ($attributename ne $::RDFSCHEMA . "ID" &&
	  $attributename ne $::RDFSCHEMA . "resource" &&
	  $attributename ne $::RDFSCHEMA . "about" &&
	  $attributename ne $::RDFSCHEMA . "aboutEach" &&
	  $attributename ne $::RDFSCHEMA . "aboutEachPrefix" &&
	  $attributename ne $::RDFSCHEMA . "bagID" &&
 	  !($attributename =~ /^xml:/)) {
	my $newElement = RDF::Element->new ("", $attributename, undef);
	my $newData    = RDF::Data->new ($attributevalue);
	$newElement->addChild ($newData);
	$self->addChild ($newElement);
	delete $self->{ATTRIBUTES}{$attributename};
      }
    }
  }

sub print
  {
    my ($self) = shift;
    my ($indent) = shift;

    printf ("%*s%s\n", $indent, " ", $self->name());

    my @children = $self->children();
    foreach my $child (@children) {
      $child->print ($indent+4);
    }
  }

package RDF::Data;

@RDF::Data::ISA = qw( RDF::Element );

sub new
  {
    my ($class) = shift;
    my ($data) = shift;
    my ($self) = $class->SUPER::new("[CDATA]");

    $self->{DATA} = $data;
    $self->{NAME} = "$data";

    bless $self, $class;
  }

sub data
  {
    my ($self) = shift;
    my ($data) = shift;

    if (defined $data) {
      $self->{DATA} = $data;
    } else {
      return $self->{DATA};
    }
  }

1;

__END__

=head1 NAME

  RDF::Parser - parse an RDF/XML document and create corresponding
  triple objects

=head1 SYNOPSIS

  use RDF::Parser;


  my $parser = new RDF::Parser;

  $parser->parsefile ($rdfdocument);

  $parser->parse ($content);

  if (my @errors = $parser->errors()) {
    foreach my $error (@errors) {
      print "Error: $error\n";
    }
  }

  my @triples = $parser->triples ();

  foreach my $triple (@triples) {
    $triple->print();
  }


  my $parser = new RDF::Parser;

  $parser->registerTellable (\&start, \&end, \&assert, \&error);

  $parser->parsefile ($rdfdocument);


  sub start  { my ($parser) = shift; }

  sub end    { my ($parser) = shift; }

  sub assert { my ($parser) = shift;
               my ($triple) = shift; }

  sub error  { my ($parser) = shift;
               my ($error)  = shift; }


  An RDF document always resides at an address; either at the
  local filesystem or somewhere on the Web addressable by a URI.

  Any local names in RDF descriptions can then be made absolute
  by applying a base URI to them.
   
  For this purpose, there is a method base() which can be used
  to set the base URI before parsing e.g.

  my $parser = new RDF::Parser;
  $parser->base ("http://www.company.com/");
  $parser->parsefile ($rdfdocument);

=head1 DESCRIPTION

  This module first spans a document tree using the Expat XML
  parser shipped with the XML::Parser Perl module. Once spanned,
  a tree traversal is activated which will create new triples
  according to the RDF Model & Syntax specification while it goes.

  The document tree consists of instances of RDF::Element and
  RDF::Data classes.

  The triples effectively describe a directed graph which consists
  of arcs and nodes. The arcs are different properties/predicates
  defined by RDF specifications or by RDF users (in RDF schemas).
  The arcs have a corresponding instances of RDF::Predicate class.

  The nodes in the RDF data model are instances of RDF::Literal
  and RDF::Resource classes. These nodes have an abstract
  base class RDF::RDFnode.

  This implementation conforms to the following specification:

  Resource Description Framework (RDF) Model and Syntax Specification -
  a W3C Recommendation dated 22 February 1999 available at:
  http://www.w3.org/TR/1999/REC-rdf-syntax-19990222/

  with the exception that the 'aboutEachPrefix' feature is not
  supported.

=head1 NAME

  RDF::Triple

=head1 SYNOPSIS

  my $triple = RDF::Triple->new ($subject, $predicate, $object);

=head1 DESCRIPTION

  The RDF::Triple class is a placeholder for Statements i.e. 3-tuples
  that consist of a subject, predicate, and an object.

=head1 NAME

  RDF::RDFnode

=head1 SYNOPSIS

  my $node = RDF::RDFnode->new ($name);

=head1 DESCRIPTION

  The RDF::RDFnode class is an abstract class that is never instantiated
  directly. It is the base class for RDF::Literal, RDF::Resource, and
  RDF::Predicate.

=head1 NAME

  RDF::Resource

=head1 SYNOPSIS

  my $resource = RDF::Resource->new ($uri);

=head1 DESCRIPTION

  The RDF::Resource class represents nodes in the RDF data model.

=head1 NAME

  RDF::Literal

=head1 SYNOPSIS

  my $literal = RDF::Literal->new ($text);

=head1 DESCRIPTION

  The RDF::Literal class represents literal values in the RDF data model.

=head1 NAME

  RDF::Predicate

=head1 SYNOPSIS

  my $predicate = RDF::Predicate->new ($name);

=head1 DESCRIPTION

  The RDF::Predicate class represents the properties in the RDF data model.

=head1 NAME

  RDF::Element

=head1 SYNOPSIS

  my $element = RDF::Element->new ($namespace, $name, \%attributelist);

=head1 DESCRIPTION

  The RDF::Element class represents XML elements that appear in an
  RDF/XML document.


=head1 NAME

  RDF::Data

=head1 SYNOPSIS

  my $data = RDF::Data->new ($pcdata);

=head1 DESCRIPTION

  The RDF::Data class represents PCDATA content in an RDF/XML document.

=head1 SEE ALSO

  XML::Parser(3)

=head1 AUTHOR

  Janne Saarela from Pro Solutions Ltd. <info@pro-solutions.com>

=cut

